/*
 * GQview
 * (C) 2004 John Ellis
 *
 * Author: John Ellis
 *
 * This software is released under the GNU General Public License (GNU GPL).
 * Please read the included file COPYING for more information.
 * This software comes with no warranty of any kind, use at your own risk!
 */


#include "gqview.h"
#include "bar_info.h"

#include "cache.h"
#include "filelist.h"
#include "info.h"
#include "utilops.h"
#include "ui_fileops.h"


#define BAR_KEYWORD_AUTOSAVE_TIME 10000


/*
 *-------------------------------------------------------------------
 * keyword / comment utils
 *-------------------------------------------------------------------
 */

gint comment_write(const gchar *path, GList *keywords, const gchar *comment)
{
	FILE *f;

	f = fopen(path, "w");
	if (!f) return FALSE;

	fprintf(f, "#GQview comment (%s)\n\n", VERSION);

	fprintf(f, "[keywords]\n");
	while (keywords)
		{
		const gchar *word = keywords->data;
		keywords = keywords->next;

		fprintf(f, "%s\n", word);
		}
	fprintf(f, "\n");

	fprintf(f, "[comment]\n");
	fprintf(f, "%s\n", (comment) ? comment : "");

	fprintf(f, "#end\n");

	fclose(f);

	return TRUE;
}

gint comment_cache_write(const gchar *path, GList *keywords, const gchar *comment)
{
	gchar *comment_dir;
	gint success = FALSE;
	mode_t mode = 0755;

	comment_dir = cache_get_location(CACHE_TYPE_METADATA, path, FALSE, &mode);

	if (cache_ensure_dir_exists(comment_dir, mode))
		{
		gchar *comment_path;
		gchar *comment_pathl;

		comment_path = g_strconcat(comment_dir, "/", filename_from_path(path),
					   GQVIEW_CACHE_EXT_METADATA, NULL);

		if (debug) printf("Saving comment: %s\n", comment_path);

		comment_pathl = path_from_utf8(comment_path);

		success = comment_write(comment_pathl, keywords, comment);

		g_free(comment_pathl);
		g_free(comment_path);
		}

	g_free(comment_dir);

	return success;
}

gint comment_read(const gchar *path, GList **keywords, gchar **comment)
{
	FILE *f;
	gchar s_buf[1024];
	gchar *key = NULL;
	GList *list = NULL;
	GString *comment_build = NULL;

	f = fopen(path, "r");
	if (!f) return FALSE;

	while (fgets(s_buf,sizeof(s_buf), f))
		{
		if (s_buf[0]=='#') continue;
		if (s_buf[0]=='[')
			{
			gint c = 0;
			gchar *ptr = s_buf + 1;

			while(ptr[c] != ']' && ptr[c] != '\n' && ptr[c] != '\0') c++;

			g_free(key);
			key = g_strndup(ptr, c);
			}
		else if (key)
			{
			gint newline = FALSE;
			gchar *ptr = s_buf;

			while (*ptr != '\n' && *ptr != '\0') ptr++;
			if (*ptr == '\n')
				{
				*ptr = '\0';
				newline = TRUE;
				}

			if (strcasecmp(key, "keywords") == 0)
				{
				if (strlen(s_buf) > 0) list = g_list_prepend(list, g_strdup(s_buf));
				}
			else if (strcasecmp(key, "comment") == 0)
				{
				if (!comment_build) comment_build = g_string_new("");
				g_string_append(comment_build, s_buf);
				if (strlen(s_buf) > 0 && newline) g_string_append_c(comment_build, '\n');
				}
			}
		}

	fclose(f);

	*keywords = g_list_reverse(list);
	if (comment_build)
		{
		if (comment) *comment = g_strdup(comment_build->str);
		g_string_free(comment_build, TRUE);
		}

	return TRUE;
}

gint comment_cache_read(const gchar *path, GList **keywords, gchar **comment)
{
	gchar *comment_dir;
	gchar *comment_path;
	gchar *comment_pathl;
	gint success = FALSE;
	mode_t mode = 0755;

	comment_dir = cache_get_location(CACHE_TYPE_METADATA, path, FALSE, &mode);
	if (!comment_dir) return FALSE;

	comment_path = g_strconcat(comment_dir, "/", filename_from_path(path),
				   GQVIEW_CACHE_EXT_METADATA, NULL);
	comment_pathl = path_from_utf8(comment_path);

	success = comment_read(comment_pathl, keywords, comment);

	g_free(comment_pathl);
	g_free(comment_path);
	g_free(comment_dir);

	return success;
}

static gchar *comment_pull(GtkWidget *textview)
{
	GtkTextBuffer *buffer;
	GtkTextIter start, end;
                                                                                                                    
	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
                                                                                                                    
	return gtk_text_buffer_get_text(buffer, &start, &end, FALSE);
}

GList *keyword_list_pull(GtkWidget *textview)
{
	GList *list = NULL;
	gchar *text;
	gchar *ptr;

	text = comment_pull(textview);

	ptr = text;
	while (*ptr != '\0')
		{
		gchar *begin;
		gint l = 0;

		while (*ptr == ' ' || *ptr == ',' || *ptr == '\n' || *ptr == '\r' || *ptr == '\b') ptr++;
		begin = ptr;
		if (*ptr != '\0')
			{
			while (*ptr != ' ' && *ptr != ',' &&
			       *ptr != '\n' && *ptr != '\r' && *ptr != '\b' &&
			       *ptr != '\0')
				{
				ptr++;
				l++;
				}
			}

		if (l > 0) list = g_list_append(list, g_strndup(begin, l));
		}

	g_free(text);

	return list;
}

void keyword_list_push(GtkWidget *textview, GList *list)
{
	GtkTextBuffer *buffer;
	GtkTextIter start, end;

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(textview));
	gtk_text_buffer_get_bounds(buffer, &start, &end);
	gtk_text_buffer_delete (buffer, &start, &end);

	while (list)
		{
		const gchar *word = list->data;
		GtkTextIter iter;

		gtk_text_buffer_get_end_iter(buffer, &iter);
		if (word) gtk_text_buffer_insert(buffer, &iter, word, -1);
		gtk_text_buffer_get_end_iter(buffer, &iter);
		gtk_text_buffer_insert(buffer, &iter, "\n", -1);

		list = list->next;
		}
}

static void metadata_set_keywords(const gchar *path, GList *list, gint add)
{
	gchar *comment = NULL;
	GList *keywords = NULL;
	GList *save_list = NULL;

	comment_cache_read(path, &keywords, &comment);

	if (add)
		{
		GList *work;

		work = list;
		while (work)
			{
			gchar *key;
			GList *p;

			key = work->data;
			work = work->next;

			p = keywords;
			while (p && key)
				{
				gchar *needle = p->data;
				p = p->next;

				if (strcmp(needle, key) == 0) key = NULL;
				}

			if (key) keywords = g_list_append(keywords, g_strdup(key));
			}
		save_list = keywords;
		}
	else
		{
		save_list = list;
		}

	comment_cache_write(path, save_list, comment);

	path_list_free(keywords);
	g_free(comment);
}

/*
 *-------------------------------------------------------------------
 * info bar
 *-------------------------------------------------------------------
 */

typedef enum {
	BAR_SORT_COPY,
	BAR_SORT_MOVE,
	BAR_SORT_LINK
} SortActionType;

typedef struct _BarInfoData BarInfoData;
struct _BarInfoData
{
	GtkWidget *vbox;
	GtkWidget *entry_file_name;
	GtkWidget *label_file_time;

	GtkWidget *keyword_view;
	GtkWidget *comment_view;

	GtkWidget *button_save;
	GtkWidget *button_set_add;
	GtkWidget *button_set_replace;

	gchar *path;

	gint changed;
	gint save_timeout_id;

	GList *(*list_func)(gpointer);
	gpointer list_data;
};

static void bar_info_write(BarInfoData *bd)
{
	GList *list;
	gchar *comment;

	if (!bd->path) return;

	list = keyword_list_pull(bd->keyword_view);
	comment = comment_pull(bd->comment_view);

	comment_cache_write(bd->path, list, comment);

	path_list_free(list);
	g_free(comment);

	bd->changed = FALSE;
	gtk_widget_set_sensitive(bd->button_save, FALSE);
}

static void bar_info_update(BarInfoData *bd)
{
	GList *keywords = NULL;
	gchar *comment = NULL;

	if (bd->entry_file_name)
		{
		gtk_entry_set_text(GTK_ENTRY(bd->entry_file_name), (bd->path) ? filename_from_path(bd->path) : "");
		}
	if (bd->label_file_time)
		{
		gtk_label_set_text(GTK_LABEL(bd->label_file_time), (bd->path) ? text_from_time(filetime(bd->path)) : "");
		}

	if (comment_cache_read(bd->path, &keywords, &comment))
		{
		keyword_list_push(bd->keyword_view, keywords);
		gtk_text_buffer_set_text(gtk_text_view_get_buffer(GTK_TEXT_VIEW(bd->comment_view)),
					 (comment) ? comment : "", -1);

		path_list_free(keywords);
		g_free(comment);
		}
	else
		{
		gtk_text_buffer_set_text(gtk_text_view_get_buffer(GTK_TEXT_VIEW(bd->keyword_view)), "", -1);
		gtk_text_buffer_set_text(gtk_text_view_get_buffer(GTK_TEXT_VIEW(bd->comment_view)), "", -1);
		}

	bd->changed = FALSE;
	gtk_widget_set_sensitive(bd->button_save, FALSE);

	gtk_widget_set_sensitive(bd->vbox, (bd->path != NULL));
}

void bar_info_set(GtkWidget *bar, const gchar *path)
{
	BarInfoData *bd;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return;

	if (bd->changed) bar_info_write(bd);

	g_free(bd->path);
	bd->path = g_strdup(path);

	bar_info_update(bd);
}

gint bar_info_event(GtkWidget *bar, GdkEvent *event)
{
	BarInfoData *bd;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return FALSE;

	if (GTK_WIDGET_HAS_FOCUS(bd->keyword_view)) return gtk_widget_event(bd->keyword_view, event);
	if (GTK_WIDGET_HAS_FOCUS(bd->comment_view)) return gtk_widget_event(bd->comment_view, event);

	return FALSE;
}

static gint bar_info_autosave(gpointer data)
{
	BarInfoData *bd = data;

	bar_info_write(bd);

	bd->save_timeout_id = -1;

	return FALSE;
}

static void bar_info_save_update(BarInfoData *bd, gint enable)
{
	if (bd->save_timeout_id != -1)
		{
		g_source_remove(bd->save_timeout_id);
		bd->save_timeout_id = -1;
		}
	if (enable)
		{
		bd->save_timeout_id = g_timeout_add(BAR_KEYWORD_AUTOSAVE_TIME, bar_info_autosave, bd);
		}
}

static void bar_info_save(GtkWidget *button, gpointer data)
{
	BarInfoData *bd = data;

	bar_info_save_update(bd, FALSE);
	bar_info_write(bd);
}

static void bar_info_set_selection(BarInfoData *bd, gint add)
{
	GList *keywords;
	GList *list = NULL;
	GList *work;

	if (!bd->list_func) return;

	keywords = keyword_list_pull(bd->keyword_view);
	if (!keywords && add) return;

	list = bd->list_func(bd->list_data);
	work = list;
	while (work)
		{
		const gchar *path = work->data;
		work = work->next;

		metadata_set_keywords(path, keywords, add);
		}

	path_list_free(list);
	path_list_free(keywords);
}

static void bar_info_set_add(GtkWidget *button, gpointer data)
{
	BarInfoData *bd = data;

	bar_info_set_selection(bd, TRUE);
}

static void bar_info_set_replace(GtkWidget *button, gpointer data)
{
	BarInfoData *bd = data;

	bar_info_set_selection(bd, FALSE);
}

static void bar_info_changed(GtkTextBuffer *buffer, gpointer data)
{
	BarInfoData *bd = data;

	bd->changed = TRUE;
	gtk_widget_set_sensitive(bd->button_save, TRUE);

	bar_info_save_update(bd, TRUE);
}

void bar_info_close(GtkWidget *bar)
{
	BarInfoData *bd;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return;

	gtk_widget_destroy(bd->vbox);
}

static void bar_info_destroy(GtkWidget *widget, gpointer data)
{
	BarInfoData *bd = data;

	if (bd->changed) bar_info_write(bd);
	bar_info_save_update(bd, FALSE);

	g_free(bd->path);

	g_free(bd);
}

GtkWidget *bar_info_new(const gchar *path, gint metadata_only)
{
	BarInfoData *bd;
	GtkWidget *frame;
	GtkWidget *table;
	GtkWidget *scrolled;
	GtkTextBuffer *buffer;
	GtkWidget *label;
	GtkWidget *tbar;

	bd = g_new0(BarInfoData, 1);

	bd->list_func = NULL;
	bd->list_data = NULL;

	bd->vbox = gtk_vbox_new(FALSE, 0);
	g_object_set_data(G_OBJECT(bd->vbox), "bar_info_data", bd);
	g_signal_connect(G_OBJECT(bd->vbox), "destroy",
			 G_CALLBACK(bar_info_destroy), bd);

	if (!metadata_only)
		{
		label = gtk_label_new(_("Image details"));
		gtk_box_pack_start(GTK_BOX(bd->vbox), label, FALSE, FALSE, 0);
		gtk_widget_show(label);
		
		bd->entry_file_name = gtk_entry_new();
		gtk_editable_set_editable(GTK_EDITABLE(bd->entry_file_name), FALSE);
		gtk_box_pack_start(GTK_BOX(bd->vbox), bd->entry_file_name, FALSE, FALSE, 0);
		gtk_widget_show(bd->entry_file_name);

		table = gtk_table_new(2, 8, FALSE);
		gtk_container_set_border_width(GTK_CONTAINER(table), 5);
		gtk_box_pack_start(GTK_BOX(bd->vbox), table, FALSE, FALSE, 0);
		gtk_widget_show(table);

		bd->label_file_time = table_add_line(table, 0, 0, _("File date:"), NULL);
		}
	else
		{
		bd->entry_file_name = NULL;
		bd->label_file_time = NULL;
		}

	frame = gtk_frame_new(_("Keywords:"));
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
	gtk_box_pack_start(GTK_BOX(bd->vbox), frame, TRUE, TRUE, 5);
	gtk_widget_show(frame);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(frame), scrolled);
	gtk_widget_show(scrolled);

	bd->keyword_view = gtk_text_view_new();
	gtk_container_add(GTK_CONTAINER(scrolled), bd->keyword_view);
	gtk_widget_show(bd->keyword_view);

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(bd->keyword_view));
	g_signal_connect(G_OBJECT(buffer), "changed",
			 G_CALLBACK(bar_info_changed), bd);

	frame = gtk_frame_new(_("Comment:"));
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
	gtk_box_pack_start(GTK_BOX(bd->vbox), frame, TRUE, TRUE, 0);
	gtk_widget_show(frame);

	scrolled = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrolled), GTK_SHADOW_IN);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				       GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_container_add(GTK_CONTAINER(frame), scrolled);
	gtk_widget_show(scrolled);

	bd->comment_view = gtk_text_view_new();
	gtk_container_add(GTK_CONTAINER(scrolled), bd->comment_view);
	gtk_widget_show(bd->comment_view);

	buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(bd->comment_view));
	g_signal_connect(G_OBJECT(buffer), "changed",
			 G_CALLBACK(bar_info_changed), bd);

#if 0
	tbar = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(bd->vbox), tbar, FALSE, FALSE, 0);
	gtk_widget_show(tbar);

	tooltips = gtk_tooltips_new();
	g_object_ref(G_OBJECT(tooltips));
	gtk_object_sink(GTK_OBJECT(tooltips));

	bd->button_set_add = layout_button(tbar, NULL, GTK_STOCK_ADD, FALSE,
				tooltips, _("Add keywords to selected files"),
				G_CALLBACK(bar_info_set_add), bd);
	bd->button_set_replace = layout_button(tbar, NULL, GTK_STOCK_CONVERT, FALSE,
				tooltips, _("Add keywords to selected files, replacing the existing ones."),
				G_CALLBACK(bar_info_set_replace), bd);
	label = gtk_hbox_new(FALSE, 0);
	gtk_box_pack_start(GTK_BOX(tbar), label, TRUE, TRUE, 0);
	gtk_widget_show(label);
	bd->button_save = layout_button(tbar, NULL, GTK_STOCK_SAVE, FALSE,
				tooltips, _("Save comment now"),
				G_CALLBACK(bar_info_save), bd);

	/* note this currently leaks the tootips widget */
#endif

	tbar = gtk_toolbar_new();
	gtk_toolbar_set_style(GTK_TOOLBAR(tbar), GTK_TOOLBAR_ICONS);
	gtk_toolbar_set_icon_size(GTK_TOOLBAR(tbar), GTK_ICON_SIZE_MENU);
	gtk_box_pack_start(GTK_BOX(bd->vbox), tbar, FALSE, FALSE, 0);
	gtk_widget_show(tbar);

	bd->button_set_add = gtk_toolbar_insert_stock(GTK_TOOLBAR(tbar), GTK_STOCK_ADD,
				_("Add keywords to selected files"), NULL,
				G_CALLBACK(bar_info_set_add), bd, -1);
	bd->button_set_replace = gtk_toolbar_insert_stock(GTK_TOOLBAR(tbar), GTK_STOCK_CONVERT,
				_("Add keywords to selected files, replacing the existing ones."), NULL,
				G_CALLBACK(bar_info_set_replace), bd, -1);
	gtk_toolbar_append_space(GTK_TOOLBAR(tbar));
	bd->button_save = gtk_toolbar_insert_stock(GTK_TOOLBAR(tbar), GTK_STOCK_SAVE,
				_("Save comment now"), NULL,
				G_CALLBACK(bar_info_save), bd, -1);

	bd->save_timeout_id = -1;

	bd->path = g_strdup(path);
	bar_info_update(bd);

	g_object_set_data(G_OBJECT(bd->vbox), "bar_info_data", bd);

	bar_info_selection(bd->vbox, 0);
	
	return bd->vbox;
}

void bar_info_set_selection_func(GtkWidget *bar, GList *(*list_func)(gpointer data), gpointer data)
{
	BarInfoData *bd;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return;

	bd->list_func = list_func;
	bd->list_data = data;
}

void bar_info_selection(GtkWidget *bar, gint count)
{
	BarInfoData *bd;
	gint enable;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return;

	enable = (count > 0 && bd->list_func != NULL);

	gtk_widget_set_sensitive(bd->button_set_add, enable);
	gtk_widget_set_sensitive(bd->button_set_replace, enable);
}

void bar_info_size_request(GtkWidget *bar, gint width)
{
	BarInfoData *bd;

	bd = g_object_get_data(G_OBJECT(bar), "bar_info_data");
	if (!bd) return;

	if (bd->entry_file_name)
		{
		gtk_widget_set_size_request(bd->entry_file_name, width, -1);
		}
}

